//////////////////////////////////////////////////////////////////////////////
//
// (C) Copyright Vicente J. Botet Escriba 2008. Distributed under the Boost
// Software License, Version 1.0. (See accompanying file
// LICENSE_1_0.txt or copy at http://www.boost.org/LICENSE_1_0.txt)
//
// See http://www.boost.org/libs/thread for documentation.
//
//////////////////////////////////////////////////////////////////////////////

#include <iostream>
#include <boost/thread/preamble.hpp>
#include <boost/thread/tss.hpp>
#include <boost/thread.hpp>

boost::mutex out_global_mutex;

void cleanup_i(boost::thread::id* data) {
	boost::mutex::scoped_lock out_guard(out_global_mutex);
	std::cout << "cleanup_i " << *data <<" " << boost::this_thread::get_id()<<std::endl;
}

boost::thread_specific_ptr<boost::thread::id> specific_ptr_i(cleanup_i);

boost::once_flag global_init_flag=BOOST_ONCE_INIT;
void global_init() {
	boost::mutex::scoped_lock out_guard(out_global_mutex);
	std::cout << "global_init" << boost::this_thread::get_id()<<std::endl;
}

void thread_init_1() {
	specific_ptr_i.reset(new boost::thread::id(boost::this_thread::get_id()));
	boost::call_once(global_init_flag, global_init);
	boost::mutex::scoped_lock out_guard(out_global_mutex);
	std::cout << "thread_init_1" << boost::this_thread::get_id()<<std::endl;
}

void thread_init_2() {
	boost::mutex::scoped_lock out_guard(out_global_mutex);
	std::cout << "thread_init_2" << boost::this_thread::get_id()<<std::endl;
}

void sleep(int sec)
{
	boost::xtime t;
	boost::xtime_get(&t,1);	t.sec += sec; boost::thread::sleep(t);
}

void run1() {
	sleep(3);
	{
		boost::mutex::scoped_lock out_guard(out_global_mutex);
		std::cout << "long operation on" << boost::this_thread::get_id()<<std::endl;
	}
	sleep(10);
}

void run2() {
	sleep(1);
	{
		boost::mutex::scoped_lock out_guard(out_global_mutex);
		std::cout << "long operation on" << boost::this_thread::get_id()<<std::endl;
	}
	sleep(20);
}

boost::thread_preamble p1(thread_init_1);
boost::thread_preamble p2(thread_init_2);


int main(int argc, char* argv[])
{

	boost::thread_preamble::proxy run1p(run1);
	boost::thread_preamble::proxy run2p(run2);
	
//	this do not compile yet
//	boost::thread pt0(boost::thread_preamble::proxy(run1));
	boost::thread pt1(run1p);
	boost::thread pt2(run2p);
	sleep(1);
	boost::thread pt12(run1p);
	sleep(1);
//	pt0.join();
	pt1.join();
	pt12.join();
//	pt2.join();
	

	return 0;
}



